package com.loafle.overflow;

import com.google.protobuf.ByteString;
import com.loafle.overflow.crawler.jmx.JmxCrawler;
import com.loafle.overflow.crawler.mongo.MongoCrawler;
import com.loafle.overflow.crawler.redis.RedisCralwer;
import com.loafle.overflow.crawler.snmp.SNMPCrawler;
import com.loafle.overflow.crawler.sql.SQLCrawler;
import com.loafle.overflow.crawler.wmi.WMICrawler;
import com.loafle.overflow.rpc.*;
import io.grpc.ServerBuilder;
import io.grpc.stub.StreamObserver;
import com.loafle.overflow.crawler.Crawler;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.lang.Boolean;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

/**
 * Created by insanity on 17. 4. 12.
 */
public class Server {

    private static final Logger logger = Logger.getLogger(Server.class.getName());
    private io.grpc.Server server;

    private static Map<String, Crawler> crawlerMap = null;

    public Server() {
        crawlerMap = new ConcurrentHashMap<>();

        addDelegate(Crawlers.SQL.name(), new SQLCrawler());
        addDelegate(Crawlers.SNMP.name(), new SNMPCrawler());
        addDelegate(Crawlers.REDIS.name(), new RedisCralwer());
        addDelegate(Crawlers.MONGODB.name(), new MongoCrawler());
        addDelegate(Crawlers.WMI.name(), new WMICrawler());
        addDelegate(Crawlers.JMX.name(), new JmxCrawler());
    }

    public int addDelegate(String name, Crawler crawler) {
        if(this.crawlerMap == null) {this.crawlerMap = new HashMap<String, Crawler>();}
        this.crawlerMap.put(name, crawler);
        return 0;
    }

    static public byte[] convertObjectToBytes(Object o) throws IOException {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        ObjectOutput out = new ObjectOutputStream(bos);
        out.writeObject(o);
        return bos.toByteArray();
    }

    public void start() throws IOException {
    /* The port on which the server should run */
        int port = 50052;
        server = ServerBuilder.forPort(port)
                .addService(new ConfigImpl())
                .addService(new DataImpl())
                .addService(new StatusImpl())
                .build()
                .start();
        logger.info("Server started, listening on " + port);
        Runtime.getRuntime().addShutdownHook(new Thread() {
            @Override
            public void run() {
                // Use stderr here since the logger may have been reset by its JVM shutdown hook.
                System.err.println("*** shutting down gRPC server since JVM is shutting down");
                Server.this.stop();
                System.err.println("*** server shut down");
            }
        });
    }

    public void stop() {
        if (server != null) {
            server.shutdown();
        }
    }

    /**
     * Await termination on the main thread since the grpc library uses daemon threads.
     */
    private void blockUntilShutdown() throws InterruptedException {
        if (server != null) {
            server.awaitTermination();
        }
    }

    static class StatusImpl extends StatusGrpc.StatusImplBase {
        @Override
        public void status(Empty request, StreamObserver<com.loafle.overflow.rpc.Boolean> responseObserver) {
            com.loafle.overflow.rpc.Boolean reply = null;
            try {
                reply = com.loafle.overflow.rpc.Boolean.newBuilder()
                        .setCheck(true)
                        .build();
                responseObserver.onNext(reply);
                responseObserver.onCompleted();
            } catch (Exception e) {
                e.printStackTrace();
            }

        }

    }


    static class DataImpl extends DataGrpc.DataImplBase {
        @Override
        public void get(Input req, StreamObserver<Output> responseObserver) {
            Crawler crawler = crawlerMap.get(req.getName().name());
            if(crawler != null) {
                try {
                    Object obj = crawler.get(req.getId());

                    Output reply = Output.newBuilder()
                            .setData(ByteString.copyFrom(convertObjectToBytes(obj)))
                            .setStartDate(333)
                            .setEndDate(333)
                            .build();
                    responseObserver.onNext(reply);
                    responseObserver.onCompleted();
                }catch(Exception e) {
                    e.printStackTrace();
                }
            }
        }
    }

    static class ConfigImpl extends ConfigGrpc.ConfigImplBase {

        @Override
        public void init(InputArray reqs, StreamObserver<Output> responseObserver) {

            List<Init> list = reqs.getInList();
            Object obj = null;
            for (Init in : list) {
                Crawler crawler = crawlerMap.get(in.getName().name());
                if (crawler != null) {
                    try {
                        obj =  crawler.init(in.getData().toByteArray());
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                }
            }

            try {
                Output reply = Output.newBuilder()
                        .setData(ByteString.copyFrom(convertObjectToBytes(obj)))
                        .setStartDate(0)
                        .setEndDate(0)
                        .build();
                responseObserver.onNext(reply);
                responseObserver.onCompleted();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        @Override
        public void add(InputAdd req, StreamObserver<Output> responseObserver) {
            Crawler crawler = crawlerMap.get(req.getName().name());
            if(crawler != null) {
                try {
                    Object obj = crawler.add(req.getData().toByteArray());
                    Output reply = Output.newBuilder()
                            .setData(ByteString.copyFrom(convertObjectToBytes(obj)))
                            .setStartDate(0)
                            .setEndDate(0)
                            .build();
                    responseObserver.onNext(reply);
                    responseObserver.onCompleted();
                }catch(Exception e) {
                    e.printStackTrace();
                }
            }
        }

        @Override
        public void remove(Input req, StreamObserver<Output> responseObserver) {
            Crawler crawler = crawlerMap.get(req.getName().name());
            if(crawler != null) {
                try {
                    Object obj = crawler.remove(req.getId());
                    Output reply = Output.newBuilder()
                            .setData(ByteString.copyFrom(convertObjectToBytes(obj)))
                            .setStartDate(0)
                            .setEndDate(0)
                            .build();
                    responseObserver.onNext(reply);
                    responseObserver.onCompleted();
                }catch(Exception e) {
                    e.printStackTrace();
                }
            }
        }

    }

    /**
     * Main launches the server from the command line.
     */
    public static void main(String[] args) throws IOException, InterruptedException {
        final Server server = new Server();
        server.start();
        server.blockUntilShutdown();
    }
}

